/* (C) 2022 YiRing, Inc. */
package com.yiring.app.service.rehearsal.impl;

import com.alibaba.fastjson.JSONArray;
import com.github.liaochong.myexcel.core.DefaultStreamExcelBuilder;
import com.yiring.app.constant.rehearsal.RehearsalPlanStatusEnum;
import com.yiring.app.domain.location.LocationLog;
import com.yiring.app.domain.location.LocationLogRepository;
import com.yiring.app.domain.location.LocationTag;
import com.yiring.app.domain.location.LocationTagRepository;
import com.yiring.app.domain.rehearsal.*;
import com.yiring.app.excel.rehearsal.RehearsalPlanExcel;
import com.yiring.app.param.rehearsal.RehearsalPlanAddParam;
import com.yiring.app.param.rehearsal.RehearsalPlanConditionParam;
import com.yiring.app.param.rehearsal.RehearsalPlanModifyParam;
import com.yiring.app.param.rehearsal.RehearsalPlanPlaybackParam;
import com.yiring.app.service.rehearsal.RehearsalPlanService;
import com.yiring.app.util.GeoUtils;
import com.yiring.app.util.JpaUtil;
import com.yiring.app.vo.rehearsal.RehearsalPlanRealTimeVo;
import com.yiring.app.vo.rehearsal.RehearsalPlanVo;
import com.yiring.app.vo.rehearsal.UserLocationVo;
import com.yiring.auth.domain.user.User;
import com.yiring.common.core.Result;
import com.yiring.common.core.Status;
import com.yiring.common.param.IdParam;
import com.yiring.common.param.PageParam;
import com.yiring.common.vo.PageVo;
import java.io.OutputStream;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;
import javax.annotation.Resource;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.Predicate;
import javax.servlet.http.HttpServletResponse;
import lombok.extern.slf4j.Slf4j;
import org.apache.poi.ss.usermodel.Workbook;
import org.locationtech.jts.geom.Geometry;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * @author tml
 * @version 1.0
 * @date 2022/5/10 14:33
 */
@Slf4j
@Service
@Transactional(rollbackFor = Exception.class)
public class RehearsalPlanServiceImpl implements RehearsalPlanService {

    @Resource
    private RehearsalPlanRepository rehearsalPlanRepository;

    @Resource
    private AccidentSpotRepository accidentSpotRepository;

    @Resource
    private EvacuationZoneRepository evacuationZoneRepository;

    @Resource
    private LocationTagRepository locationTagRepository;

    @Resource
    private LocationLogRepository locationLogRepository;

    @Resource
    private RehearsalPlanLogRepository rehearsalPlanLogRepository;

    @Override
    public Result<String> addOne(RehearsalPlanAddParam param) {
        Result<String> result = checkEnable(param.getAccidentSpotId(), param.getEvacuationZoneId());
        if (Status.OK.value() != result.getStatus()) {
            return result;
        }
        RehearsalPlan rehearsalPlan = param.transform();
        rehearsalPlanRepository.saveAndFlush(rehearsalPlan);
        return Result.ok();
    }

    @Override
    public Result<String> modifyOne(RehearsalPlanModifyParam param) {
        Result<String> result = checkEnable(param.getAccidentSpotId(), param.getEvacuationZoneId());
        if (Status.OK.value() != result.getStatus()) {
            return result;
        }
        Optional<RehearsalPlan> optional = rehearsalPlanRepository.findById(param.getId());
        if (optional.isEmpty()) {
            return Result.no(Status.EXPECTATION_FAILED, "要修改的演练计划不存在");
        }
        //只有未开始的演练计划才能修改
        if (!RehearsalPlanStatusEnum.UN_START.getCode().equals(optional.get().getStatus())) {
            return Result.no(Status.EXPECTATION_FAILED, "只有未开始的演练计划才能修改");
        }
        JpaUtil.copyNotNullProperties(param.transform(), optional.get());
        return Result.ok();
    }

    private Result<String> checkEnable(Long accidentSpotId, Long evacuationZoneId) {
        Optional<AccidentSpot> spotOptional = accidentSpotRepository.findById(accidentSpotId);
        if (spotOptional.isEmpty()) {
            return Result.no(Status.EXPECTATION_FAILED, "事故点不存在");
        }
        if (!spotOptional.get().getEnable()) {
            return Result.no(Status.EXPECTATION_FAILED, "事故点未启用");
        }
        Optional<EvacuationZone> zoneOptional = evacuationZoneRepository.findById(evacuationZoneId);
        if (zoneOptional.isEmpty()) {
            return Result.no(Status.EXPECTATION_FAILED, "撤离区不存在");
        }
        if (!zoneOptional.get().getEnable()) {
            return Result.no(Status.EXPECTATION_FAILED, "撤离区未启用");
        }
        return Result.ok();
    }

    @Override
    public Result<String> removeOne(IdParam idParam) {
        Optional<RehearsalPlan> optional = rehearsalPlanRepository.findById(idParam.getId());
        if (optional.isEmpty()) {
            return Result.no(Status.EXPECTATION_FAILED, "要删除的演练计划不存在");
        }
        //只有未开始的演练计划才能删除
        if (!RehearsalPlanStatusEnum.UN_START.getCode().equals(optional.get().getStatus())) {
            return Result.no(Status.EXPECTATION_FAILED, "只有未开始的演练计划才能删除");
        }
        optional.get().setDeleted(true);
        return Result.ok();
    }

    @Override
    public Result<RehearsalPlanVo> findOne(IdParam idParam) {
        Optional<RehearsalPlan> optional = rehearsalPlanRepository.findById(idParam.getId());
        if (optional.isEmpty()) {
            return Result.no(Status.EXPECTATION_FAILED, "要删除的演练计划不存在");
        }
        RehearsalPlanVo planVo = RehearsalPlanVo.transform(optional.get());
        return Result.ok(planVo);
    }

    @Override
    public Result<PageVo<RehearsalPlanVo>> findList(RehearsalPlanConditionParam conditionParam, PageParam pageParam) {
        Pageable pageable = PageParam.toPageable(pageParam);
        Specification<RehearsalPlan> specification = getSpecification(conditionParam);
        Page<RehearsalPlan> page = rehearsalPlanRepository.findAll(specification, pageable);
        List<RehearsalPlanVo> voList = page.get().map(RehearsalPlanVo::transform).collect(Collectors.toList());
        PageVo<RehearsalPlanVo> pageVo = PageVo.build(voList, voList.size());
        return Result.ok(pageVo);
    }

    @Override
    public void export(RehearsalPlanConditionParam conditionParam, HttpServletResponse response) {
        Specification<RehearsalPlan> specification = getSpecification(conditionParam);
        List<RehearsalPlan> page = rehearsalPlanRepository.findAll(specification);
        List<RehearsalPlanExcel> list = page.stream().map(RehearsalPlanExcel::transform).collect(Collectors.toList());
        try (
            DefaultStreamExcelBuilder<RehearsalPlanExcel> streamExcelBuilder = DefaultStreamExcelBuilder
                .of(RehearsalPlanExcel.class)
                .threadPool(Executors.newFixedThreadPool(2))
                .rowHeight(14)
                .titleRowHeight(14)
                .style(
                    "cell->vertical-align:center;text-align:center",
                    "title->vertical-align:center;text-align:center;font-weight:bold;font-family:等线"
                )
                .start()
        ) {
            streamExcelBuilder.append(list);

            String fileName = URLEncoder.encode("演练计划列表.xlsx", StandardCharsets.UTF_8);
            response.setContentType("application/octet-stream");
            response.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
            response.setHeader("Content-Disposition", "attachment;filename=" + fileName);
            OutputStream out = response.getOutputStream();

            Workbook workbook = streamExcelBuilder.fixedTitles().build();
            workbook.write(out);
            workbook.close();
            out.flush();
            out.close();
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            throw new RuntimeException("报警类型列表导出失败: " + e.getMessage());
        }
    }

    private Specification<RehearsalPlan> getSpecification(RehearsalPlanConditionParam conditionParam) {
        return (root, query, criteriaBuilder) -> {
            ArrayList<Predicate> list = new ArrayList<>();
            if (conditionParam.getStatus() != null) {
                list.add(criteriaBuilder.equal(root.get(RehearsalPlan.Fields.status), conditionParam.getStatus()));
            }
            if (conditionParam.getRiskLevel() != null) {
                list.add(
                    criteriaBuilder.equal(root.get(RehearsalPlan.Fields.riskLevel), conditionParam.getRiskLevel())
                );
            }
            if (conditionParam.getBeginTime() != null) {
                list.add(
                    criteriaBuilder.greaterThanOrEqualTo(
                        root.get(RehearsalPlan.Fields.rehearsalTime),
                        conditionParam.getBeginTime()
                    )
                );
            }
            if (conditionParam.getEndTime() != null) {
                list.add(
                    criteriaBuilder.lessThanOrEqualTo(
                        root.get(RehearsalPlan.Fields.rehearsalTime),
                        conditionParam.getEndTime()
                    )
                );
            }
            Predicate[] array = list.toArray(new Predicate[0]);
            return criteriaBuilder.and(array);
        };
    }

    @Override
    public Result<String> start(IdParam idParam) {
        Optional<RehearsalPlan> optional = rehearsalPlanRepository.findById(idParam.getId());
        if (optional.isEmpty()) {
            return Result.no(Status.EXPECTATION_FAILED, "要启动的演练计划不存在");
        }
        RehearsalPlan rehearsalPlan = optional.get();
        if (!RehearsalPlanStatusEnum.UN_START.getCode().equals(rehearsalPlan.getStatus())) {
            return Result.no(Status.EXPECTATION_FAILED, "演练计划已经启动过了");
        }
        rehearsalPlan.setStatus(RehearsalPlanStatusEnum.RUNNING.getCode());
        //事故点信息
        AccidentSpot accidentSpot = rehearsalPlan.getAccidentSpot();
        Integer scope = rehearsalPlan.getScope();
        //计算影响的范围
        Geometry geometry = GeoUtils.expandGeometry(accidentSpot.getGeometry(), scope);
        //获取在事故点及周边的员工
        List<LocationTag> inArea = locationTagRepository.findInArea(geometry);

        //演练计划实施数据落库
        List<Long> idList = inArea.stream().map(e -> e.getUser().getId()).collect(Collectors.toList());
        String ids = JSONArray.toJSONString(idList);
        RehearsalPlanLog rehearsalPlanLog = RehearsalPlanLog
            .builder()
            .rehearsalPlan(rehearsalPlan)
            .beginTime(LocalDateTime.now())
            .accidentPeopleNum(inArea.size())
            .userIds(ids)
            .geometry(geometry)
            .build();
        rehearsalPlanLogRepository.saveAndFlush(rehearsalPlanLog);

        //todo:给这些标签卡绑定的用户发送 App 消息推送 - 同时是否会生成广播消息未知？

        return Result.ok();
    }

    @Override
    public Result<String> stop(IdParam idParam) {
        Optional<RehearsalPlan> optional = rehearsalPlanRepository.findById(idParam.getId());
        if (optional.isEmpty()) {
            return Result.no(Status.EXPECTATION_FAILED, "要启动的演练计划不存在");
        }
        RehearsalPlan rehearsalPlan = optional.get();
        if (!RehearsalPlanStatusEnum.RUNNING.getCode().equals(rehearsalPlan.getStatus())) {
            return Result.no(Status.EXPECTATION_FAILED, "演练计划没有在进行中");
        }
        rehearsalPlan.setStatus(RehearsalPlanStatusEnum.END.getCode());
        RehearsalPlanLog rehearsalPlanLog = rehearsalPlanLogRepository.findByRehearsalPlan(rehearsalPlan);

        //查询出结束时在撤离区的员工,并落库
        EvacuationZone evacuationZone = rehearsalPlan.getEvacuationZone();
        int inAreaNum = locationTagRepository.findInAreaNum(evacuationZone.getGeometry());
        rehearsalPlanLog.setEndTime(LocalDateTime.now());
        rehearsalPlanLog.setEvacuationPeopleNum(inAreaNum);
        return Result.ok();
    }

    @Override
    public Result<RehearsalPlanRealTimeVo> findRealTime(IdParam idParam) {
        RehearsalPlan rehearsalPlan = RehearsalPlan.builder().id(idParam.getId()).build();
        RehearsalPlanLog rehearsalPlanLog = rehearsalPlanLogRepository.findByRehearsalPlan(rehearsalPlan);
        if (rehearsalPlanLog == null) {
            return Result.no(Status.EXPECTATION_FAILED, "演练计划不存在或未启动");
        }
        if (rehearsalPlanLog.getEndTime() != null) {
            return Result.no(Status.EXPECTATION_FAILED, "演练计划已停止");
        }
        String idString = rehearsalPlanLog.getUserIds();
        List<Long> idList = JSONArray.parseArray(idString, Long.class);
        //查看员工位置
        List<LocationTag> locationTagList = locationTagRepository.findByUserIds(idList);
        List<UserLocationVo> userLocationVos = locationTagList
            .stream()
            .map(e -> new UserLocationVo(e.getUser().getId(), e.getUser().getRealName(), e.getPoint()))
            .collect(Collectors.toList());
        //查看撤离区人员数量
        Geometry geometry = rehearsalPlanLog.getRehearsalPlan().getEvacuationZone().getGeometry();
        int inAreaNum = locationTagRepository.findInAreaNum(geometry);

        RehearsalPlanRealTimeVo vo = RehearsalPlanRealTimeVo
            .builder()
            .beginPeopleNum(rehearsalPlanLog.getAccidentPeopleNum())
            .evacuationPeopleNum(inAreaNum)
            .userLocation(userLocationVos)
            .time(LocalDateTime.now())
            .build();
        return Result.ok(vo);
    }

    @Override
    public Result<PageVo<RehearsalPlanRealTimeVo>> findPlayback(RehearsalPlanPlaybackParam param) {
        RehearsalPlan rehearsalPlan = RehearsalPlan.builder().id(param.getId()).build();
        RehearsalPlanLog rehearsalPlanLog = rehearsalPlanLogRepository.findByRehearsalPlan(rehearsalPlan);
        //校验参数
        Result<PageVo<RehearsalPlanRealTimeVo>> result = checkPlaybackParam(param, rehearsalPlanLog);
        if (Status.OK.value() != result.getStatus()) {
            return result;
        }
        //撤离区空间信息
        Geometry evacuationGeometry = rehearsalPlanLog.getRehearsalPlan().getEvacuationZone().getGeometry();
        String idString = rehearsalPlanLog.getUserIds();
        List<Long> idList = JSONArray.parseArray(idString, Long.class);
        //初始化查询的结束时间
        LocalDateTime endTime;
        if (param.getBeginTime().plusMinutes(1).isAfter(rehearsalPlanLog.getEndTime())) {
            endTime = rehearsalPlanLog.getEndTime();
        } else {
            endTime = param.getBeginTime().plusMinutes(1);
        }
        //查询开始时间到一分钟后这段时间的员工定位信息
        Specification<LocationLog> specification = (root, query, criteriaBuilder) -> {
            ArrayList<Predicate> list = new ArrayList<>();
            CriteriaBuilder.In<Object> in = criteriaBuilder.in(root.get(LocationLog.Fields.user));
            for (Long id : idList) {
                in.value(id);
            }
            list.add(in);
            list.add(
                criteriaBuilder.greaterThanOrEqualTo(root.get(LocationLog.Fields.locationTime), param.getBeginTime())
            );
            list.add(criteriaBuilder.lessThanOrEqualTo(root.get(LocationLog.Fields.locationTime), endTime));
            Predicate[] array = list.toArray(new Predicate[0]);
            return criteriaBuilder.and(array);
        };
        List<LocationLog> locationLogList = locationLogRepository.findAll(specification);

        //处理每秒中员工的定位信息，并包装
        ArrayList<RehearsalPlanRealTimeVo> voList = new ArrayList<>();
        HashSet<Long> set = new HashSet<>();
        //每次筛选的开始时间和结束时间(一秒间隔)
        LocalDateTime tabTime = param.getBeginTime();
        LocalDateTime tabTimeOnes = tabTime.plusSeconds(1);
        while (tabTime.isBefore(endTime)) {
            ArrayList<UserLocationVo> userLocationVos = new ArrayList<>();
            RehearsalPlanRealTimeVo realTimeVo = RehearsalPlanRealTimeVo
                .builder()
                .beginPeopleNum(rehearsalPlanLog.getAccidentPeopleNum())
                .evacuationPeopleNum(0)
                .userLocation(userLocationVos)
                .time(tabTime)
                .build();
            //循环员工的信息，筛选在这一秒的定位信息
            for (LocationLog locationLog : locationLogList) {
                LocalDateTime locationTime = locationLog.getLocationTime();
                if (tabTime.isBefore(locationTime) && tabTimeOnes.isAfter(locationTime)) {
                    User user = locationLog.getUser();
                    //同一秒只记录用户的一次定位信息
                    if (!set.contains(user.getId())) {
                        set.add(user.getId());
                        UserLocationVo userLocationVo = UserLocationVo
                            .builder()
                            .userId(user.getId())
                            .realmName(user.getRealName())
                            .point(locationLog.getPoint())
                            .build();
                        userLocationVos.add(userLocationVo);
                        //如果是这一秒内的定位信息，记录已在撤离区的人数
                        boolean contains = evacuationGeometry.contains(locationLog.getPoint());
                        if (contains) {
                            realTimeVo.setEvacuationPeopleNum(realTimeVo.getEvacuationPeopleNum() + 1);
                        }
                    }
                }
            }
            //将这一秒的数据添加到结果集
            voList.add(realTimeVo);
            //用户去重的set清空
            set.clear();
            //设置下一秒
            tabTime = tabTime.plusSeconds(1);
            tabTimeOnes = tabTime.plusSeconds(1);
        }
        PageVo<RehearsalPlanRealTimeVo> pageVo = PageVo.build(voList, voList.size());
        return Result.ok(pageVo);
    }

    private Result<PageVo<RehearsalPlanRealTimeVo>> checkPlaybackParam(
        RehearsalPlanPlaybackParam param,
        RehearsalPlanLog rehearsalPlanLog
    ) {
        if (rehearsalPlanLog == null) {
            return Result.no(Status.EXPECTATION_FAILED, "演练计划不存在或未启动");
        }
        if (rehearsalPlanLog.getEndTime() == null) {
            return Result.no(Status.EXPECTATION_FAILED, "演练计划停止后才能生成演练回放");
        }
        if (param.getBeginTime() == null) {
            param.setBeginTime(rehearsalPlanLog.getBeginTime());
        }
        if (param.getBeginTime().isBefore(rehearsalPlanLog.getBeginTime())) {
            return Result.no(Status.EXPECTATION_FAILED, "演练回放只能回放演练计划开始之后的数据");
        }
        return Result.ok();
    }
}
