/* (C) 2022 YiRing, Inc. */
package com.yiring.auth.service.user.impl;

import cn.dev33.satoken.secure.SaSecureUtil;
import cn.hutool.core.collection.ListUtil;
import cn.hutool.core.convert.Convert;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import com.github.liaochong.myexcel.core.DefaultStreamExcelBuilder;
import com.yiring.auth.domain.user.User;
import com.yiring.auth.domain.user.UserRepository;
import com.yiring.auth.excel.user.UserExportExcel;
import com.yiring.auth.param.user.UserExportParam;
import com.yiring.auth.param.user.UserFindParam;
import com.yiring.auth.param.user.UserSaveParam;
import com.yiring.auth.service.user.UserService;
import com.yiring.auth.vo.user.UserVo;
import com.yiring.common.core.Result;
import com.yiring.common.core.Status;
import com.yiring.common.domain.BasicEntity;
import com.yiring.common.param.IdParam;
import com.yiring.common.param.PageParam;
import com.yiring.common.vo.PageVo;
import java.io.OutputStream;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;
import javax.annotation.Resource;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.Order;
import javax.persistence.criteria.Predicate;
import javax.servlet.http.HttpServletResponse;
import lombok.extern.slf4j.Slf4j;
import org.apache.poi.ss.usermodel.Workbook;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * 部门信息控制器
 *
 * @author LJ-2204
 * @date 2022/5/10
 */

@Transactional(rollbackFor = RuntimeException.class)
@Service
@Slf4j
public class UserServiceImpl implements UserService {

    @Resource
    UserRepository userRepository;

    @Override
    public Result<String> add(UserSaveParam param) {
        userRepository.save(UserSaveParam.transform(param));
        return Result.ok();
    }

    @Override
    public Result<PageVo<UserVo>> page(UserFindParam param, PageParam pageParam) {
        Specification<User> specification = (root, cq, cb) -> {
            List<Predicate> predicates = ListUtil.toList();

            if (StrUtil.isNotEmpty(param.getRealName())) {
                predicates.add(cb.like(root.get(User.Fields.realName), "%" + param.getRealName() + "%"));
            }

            if (ObjectUtil.isNotEmpty(param.getDeptId())) {
                predicates.add(
                    cb.equal(root.get(User.Fields.department).get(BasicEntity.Fields.id), param.getDeptId())
                );
            }

            if (ObjectUtil.isNotEmpty(param.getRoles())) {
                CriteriaBuilder.In<Long> in = cb.in(root.join(User.Fields.roles).get(BasicEntity.Fields.id));
                String[] split = param.getRoles().split(",");
                Long[] roles = Convert.toLongArray(split);
                for (Long roleId : roles) {
                    in.value(roleId);
                }
                predicates.add(cb.and(in));
            }

            Order order = cb.desc(root.get(BasicEntity.Fields.createTime));

            return cq.orderBy(order).where(predicates.toArray(new Predicate[0])).getRestriction();
        };

        // 分页
        Pageable pageable = PageRequest.of(pageParam.getPageNo() - 1, pageParam.getPageSize());
        Page<User> users = userRepository.findAll(specification, pageable);
        List<UserVo> userVos = users.stream().map(UserVo::transform).collect(Collectors.toList());
        PageVo<UserVo> pageVo = PageVo.build(userVos, users.getTotalElements());
        return Result.ok(pageVo);
    }

    @Override
    public Result<String> reset(IdParam idParam) {
        Optional<User> userOptional = userRepository.findById(idParam.getId());

        if (userOptional.isPresent()) {
            User user = userOptional.get();
            user.setPassword(SaSecureUtil.sha256("123456"));
            userRepository.save(user);
            return Result.ok();
        }

        return Result.no(Status.BAD_REQUEST);
    }

    @Override
    public Result<String> upStatus(IdParam idParam) {
        Optional<User> userOptional = userRepository.findById(idParam.getId());

        if (userOptional.isPresent()) {
            User user = userOptional.get();
            if (ObjectUtil.isEmpty(user.getEnabled())) {
                user.setEnabled(true);
            } else {
                user.setEnabled(!user.getEnabled());
            }
            userRepository.save(user);
            return Result.ok(user.getEnabled().toString());
        }

        return Result.no(Status.BAD_REQUEST);
    }

    @Override
    public void export(UserExportParam param, HttpServletResponse response) {
        Specification<User> specification = (root, cq, cb) -> {
            List<Predicate> predicates = ListUtil.toList();

            if (StrUtil.isNotEmpty(param.getRealName())) {
                predicates.add(cb.like(root.get(User.Fields.realName), "%" + param.getRealName() + "%"));
            }

            if (ObjectUtil.isNotEmpty(param.getDeptId())) {
                predicates.add(
                    cb.equal(root.get(User.Fields.department).get(BasicEntity.Fields.id), param.getDeptId())
                );
            }

            if (ObjectUtil.isNotEmpty(param.getRoles())) {
                CriteriaBuilder.In<Long> in = cb.in(root.join(User.Fields.roles).get(BasicEntity.Fields.id));
                String[] split = param.getRoles().split(",");
                Long[] roles = Convert.toLongArray(split);
                for (Long roleId : roles) {
                    in.value(roleId);
                }
                predicates.add(cb.and(in));
            }

            Order order = cb.desc(root.get(BasicEntity.Fields.createTime));

            return cq.orderBy(order).where(predicates.toArray(new Predicate[0])).getRestriction();
        };

        List<User> users = userRepository.findAll(specification);
        List<UserExportExcel> userExportExcels = users
            .stream()
            .map(UserExportExcel::transform)
            .collect(Collectors.toList());
        try (
            DefaultStreamExcelBuilder<UserExportExcel> streamExcelBuilder = DefaultStreamExcelBuilder
                .of(UserExportExcel.class)
                .threadPool(Executors.newFixedThreadPool(2))
                .rowHeight(14)
                .titleRowHeight(14)
                .style(
                    "cell->vertical-align:center;text-align:center",
                    "title->vertical-align:center;text-align:center;font-weight:bold;font-family:等线"
                )
                .start()
        ) {
            streamExcelBuilder.append(userExportExcels);

            String fileName = URLEncoder.encode("用户信息.xlsx", StandardCharsets.UTF_8);

            response.setContentType("application/octet-stream");
            response.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
            response.setHeader("Content-Disposition", "attachment;filename=" + fileName);

            OutputStream out = response.getOutputStream();

            Workbook workbook = streamExcelBuilder.fixedTitles().build();
            workbook.write(out);
            workbook.close();
            out.flush();
            out.close();
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            throw new RuntimeException("导出用户信息失败: " + e.getMessage());
        }
    }
}
